import json
import logging
import os
import tempfile
from pathlib import Path

from firecloud_api_cds import api as firecloud_api
from firecloud_api_cds.api import __post as firecloud_post

from nebelung.types import PersistedWdl
from nebelung.utils import call_firecloud_api
from nebelung.wdl import GistedWdl


class TerraWorkflow:
    def __init__(
        self,
        method_namespace: str,
        method_name: str,
        method_config_namespace: str,
        method_config_name: str,
        method_synopsis: str,
        workflow_wdl_path: Path,
        method_config_json_path: Path,
        github_pat: str | None = None,
        womtool_jar: str | None = None,
    ) -> None:
        self.method_namespace = method_namespace
        self.method_name = method_name
        self.method_config_namespace = method_config_namespace
        self.method_config_name = method_config_name
        self.method_synopsis = method_synopsis
        self.workflow_wdl_path = workflow_wdl_path
        self.method_config_json_path = method_config_json_path
        self.github_pat = os.getenv("GITHUB_PAT", github_pat)
        self.womtool_jar = os.getenv("WOMTOOL_JAR", womtool_jar)

        self.method_config = json.load(open(self.method_config_json_path, "r"))
        self.persisted_wdl_script: PersistedWdl | None = None

    def persist_method_on_github(self) -> None:
        """
        Upload the method's WDL script to GitHub, rewriting import statements for
        dependent WDL scripts as needed.
        """

        assert self.github_pat is not None, (
            "A GitHub personal access token must be defined to persist this method on "
            "GitHub. Set the GITHUB_PAT environment variable or the `github_pat` "
            "argument when instantiating this `TerraWorkflow` instance."
        )

        assert self.womtool_jar is not None, (
            "A path to a WOMTool .jar file is required to validate the WDL script. Set "
            "the WOMTOOL_JAR environment variable or the `womtool_jar` argument when "
            "instantiating this `TerraWorkflow` instance."
        )

        if self.persisted_wdl_script is None:
            logging.info(f"Persisting {self.workflow_wdl_path} on GitHub")
            gisted_wdl = GistedWdl(
                method_name=self.method_name,
                github_pat=self.github_pat,
                womtool_jar=self.womtool_jar,
            )
            self.persisted_wdl_script = gisted_wdl.persist_wdl_script(
                wdl_path=self.workflow_wdl_path
            )

    def update_method(self, owners: list[str]) -> dict:
        """
        Update a Firecloud method.

        :param owners: a list of Firecloud users/groups to set as owners
        :return: the latest method's snapshot
        """

        # get contents of WDL uploaded to GCS
        self.persist_method_on_github()
        assert self.persisted_wdl_script is not None

        with tempfile.NamedTemporaryFile("w") as f:
            f.write(self.persisted_wdl_script["wdl"])
            f.flush()

            logging.info(f"Updating method {self.method_name}")
            snapshot = call_firecloud_api(
                firecloud_api.update_repository_method,
                namespace=self.method_namespace,
                method=self.method_name,
                synopsis=self.method_synopsis,
                wdl=f.name,
            )

        logging.info(f"Setting method {self.method_name} ACL")
        call_firecloud_api(
            firecloud_api.update_repository_method_acl,
            namespace=self.method_namespace,
            method=self.method_name,
            snapshot_id=snapshot["snapshotId"],
            acl_updates=[{"user": x, "role": "OWNER"} for x in owners],
        )

        logging.info(f"Setting method repository config ACL in {self.method_namespace}")
        # the firecloud package doesn't have a wrapper for this endpoint
        call_firecloud_api(
            firecloud_post,
            methcall=f"configurations/{self.method_namespace}/permissions",
            json=[{"user": x, "role": "OWNER"} for x in owners],
        )

        return snapshot

    def get_method_snapshots(self) -> list[dict]:
        """
        Get all of the snapshots of the method.

        :return: list of snapshot information, most recent first
        """

        logging.info(f"Getting {self.method_name} method snapshots")
        snapshots = call_firecloud_api(
            firecloud_api.list_repository_methods,
            namespace=self.method_namespace,
            name=self.method_name,
        )

        snapshots.sort(key=lambda x: x["snapshotId"], reverse=True)
        return snapshots

    def delete_old_method_snapshots(self, n_snapshots_to_keep: int) -> None:
        """
        Delete all but `n_snapshots_to_keep` of the most recent snapshots of the method.

        :param n_snapshots_to_keep: the number of snapshots to keep
        """

        snapshots = self.get_method_snapshots()

        to_delete = snapshots[n_snapshots_to_keep:]
        logging.info(f"Deleting {len(to_delete)} old {self.method_name} snapshot(s)")

        for x in to_delete:
            call_firecloud_api(
                firecloud_api.delete_repository_method,
                namespace=self.method_namespace,
                name=self.method_name,
                snapshot_id=x["snapshotId"],
            )



from typing import Any, NotRequired, Optional, TypedDict, TypeVar

import pandas as pd
import pandera as pa
import pandera.typing
from pandera.api.pandas.model_config import BaseConfig as PaBaseConfig
from pandera.typing import Series
from pydantic import BaseModel


class PersistedWdl(TypedDict):
    wdl: str
    public_url: str


class SubmittableEntities(TypedDict):
    unsubmitted: set[str]
    running: set[str]
    retryable: set[str]
    failed: set[str]


class TerraJobSubmissionKwargs(TypedDict):
    entity: NotRequired[str | None]
    etype: NotRequired[str | None]
    expression: NotRequired[str | None]
    use_callcache: NotRequired[bool | None]
    delete_intermediate_output_files: NotRequired[bool | None]
    use_reference_disks: NotRequired[bool | None]
    memory_retry_multiplier: NotRequired[float | None]
    workflow_failure_mode: NotRequired[str | None]
    user_comment: NotRequired[str | None]


class TaskResult(BaseModel):
    completed_at: Optional[Any] = None
    crc32c_hash: Optional[str] = None
    format: Optional[str] = None
    id: Optional[str] = None
    label: Optional[str] = None
    size: Optional[int] = None
    task_entity_id: Optional[int] = None
    terra_entity_name: Optional[str] = None
    terra_entity_type: Optional[str] = None
    terra_method_config_name: Optional[str] = None
    terra_method_config_namespace: Optional[str] = None
    terra_submission_id: Optional[str] = None
    terra_workflow_id: Optional[str] = None
    terra_workflow_inputs: Optional[dict] = None
    terra_workflow_root_dir: Optional[str] = None
    terra_workspace_id: Optional[str] = None
    terra_workspace_name: Optional[str] = None
    terra_workspace_namespace: Optional[str] = None
    url: Optional[str] = None
    value: Optional[dict] = None
    workflow_name: Optional[str] = None


class CoercedDataFrame(pa.DataFrameModel):
    class Config(PaBaseConfig):
        coerce = True  # convert to indicated dtype upon TypedDataFrame init


class Submissions(CoercedDataFrame):
    deleteIntermediateOutputFiles: Series[pd.BooleanDtype]
    methodConfigurationDeleted: Series[pd.BooleanDtype]
    methodConfigurationName: Series[pd.StringDtype]
    methodConfigurationNamespace: Series[pd.StringDtype]
    status: Series[pd.StringDtype]
    submissionDate: Series[pd.Timestamp]
    submissionId: Series[pd.StringDtype]
    submissionRoot: Series[pd.StringDtype]
    submitter: Series[pd.StringDtype]
    useCallCache: Series[pd.BooleanDtype]
    userComment: Series[pd.StringDtype] = pa.Field(nullable=True)


class SubmittedEntities(CoercedDataFrame):
    entity_type: Series[pd.StringDtype]
    entity_id: Series[pd.StringDtype]
    status: Series[pd.StringDtype] = pa.Field(
        isin=[
            "Queued",
            "Submitted",
            "Launching",
            "Running",
            "Aborted",
            "Aborting",
            "Succeeded",
            "Failed",
        ]
    )


PanderaBaseSchema = TypeVar("PanderaBaseSchema", bound=CoercedDataFrame)
TypedDataFrame = pandera.typing.DataFrame
T = TypeVar("T")



import os
import re
import subprocess
import tempfile
from pathlib import Path

import github

from nebelung.types import PersistedWdl

IMPORT_PATTERN = re.compile(r"^import\s+\"(?!http)([^\"]+)\"\s+as\s+(\S+)")


class GistedWdl:
    def __init__(self, method_name: str, github_pat: str, womtool_jar: str):
        self.method_name = method_name
        self.github_pat = github_pat
        self.womtool_jar = womtool_jar
        self.gist = None

    def find_or_create_gist(self):
        """
        Find or create a GitHub gist for a Terra method.
        """

        github_auth = github.Auth.Token(self.github_pat)
        g = github.Github(auth=github_auth)

        # check if there's already a gist for this method by comparing the descriptions
        gists = g.get_user().get_gists()
        matching_gists = [x for x in gists if x.description == self.method_name]

        if len(matching_gists) == 0:
            self.gist = g.get_user().create_gist(  # pyright: ignore
                description=self.method_name,
                files={
                    "README.md": github.InputFileContent(
                        content=f"Persisted WDL for the `{self.method_name}` method"
                    )
                },
                public=True,
            )
        else:
            self.gist = matching_gists[0]

    def persist_wdl_script(self, wdl_path: Path) -> PersistedWdl:
        """
        Save a copy of a local WDL script and its dependencies to GitHub, rewriting the
        import statements to refer to the public gist.github.com URLs.

        :param wdl_path: the absolute path to a WDL script
        :return: a dictionary of the uploaded WDL, its public URL, and its
        `pipeline_version` (if found as a variable in the `workflow`)
        """

        with open(wdl_path, "r") as f:
            wdl_lines = f.readlines()

        wdl_basename = os.path.basename(wdl_path)
        buffer = []  # build buffer of lines in the WDL file

        for line in wdl_lines:
            if import_match := re.match(IMPORT_PATTERN, line):
                # need to upload the dependent WDL file to GCS and rewrite the `import`
                # statement in this WDL file to the dependent file's public GCS URL

                rel_path = import_match[1]  # relative path to the dependent file
                import_alias = import_match[2]  # `as <x>` component of the statement

                # absolute path to the dependent file
                abs_wdl_path = Path.joinpath(wdl_path.parent.absolute(), rel_path)

                # recurse: upload the dependent WDL and get its URL
                imported_public_url = self.persist_wdl_script(abs_wdl_path)[
                    "public_url"
                ]

                # replace the local relative import with the absolute one
                converted_line = f'import "{imported_public_url}" as {import_alias}'
                buffer.append(converted_line)

            else:
                # keep writing to buffer
                buffer.append(line.rstrip())

        # construct the final version of this WDL script
        converted_wdl = "\n".join(buffer)

        # validate the WDL before uploading
        with tempfile.NamedTemporaryFile("w") as f:
            f.write(converted_wdl)
            f.flush()

            res = subprocess.run(
                ["java", "-jar", self.womtool_jar, "validate", f.name],
                capture_output=True,
            )

            if res.returncode != 0 or "Success" not in res.stdout.decode():
                raise ChildProcessError(
                    f"Error validating {wdl_path}: {res.stderr.decode()}"
                )

        # update the gist with the new version of the WDL script
        if self.gist is None:
            self.find_or_create_gist()

        assert self.gist is not None

        self.gist.edit(
            files={wdl_basename: github.InputFileContent(content=converted_wdl)}
        )

        # return the public URL of the new version of the WDL script so that it can be
        # imported by its parent (if there is one) or used by the calling function
        raw_url = self.gist.files[wdl_basename].raw_data["raw_url"]

        return {"wdl": converted_wdl, "public_url": raw_url}



import datetime
import logging
import pathlib
from io import StringIO
from typing import Any, Iterable, Type, Unpack

import pandas as pd
from firecloud_api_cds import api as firecloud_api
from pd_flatten import pd_flatten

from nebelung.terra_workflow import TerraWorkflow
from nebelung.types import (
    PanderaBaseSchema,
    Submissions,
    SubmittableEntities,
    SubmittedEntities,
    TaskResult,
    TerraJobSubmissionKwargs,
    TypedDataFrame,
)
from nebelung.utils import batch_evenly, call_firecloud_api, type_data_frame


class TerraWorkspace:
    def __init__(
        self,
        workspace_namespace: str,
        workspace_name: str,
        owners: list[str] | None = None,
    ) -> None:
        self.workspace_namespace = workspace_namespace
        self.workspace_name = workspace_name
        self.owners = [] if owners is None else owners

    def get_entities(
        self,
        entity_type: str,
        pandera_schema: Type[PanderaBaseSchema] | None = None,
    ) -> pd.DataFrame | TypedDataFrame[PanderaBaseSchema]:
        """
        Get a data frame of entities from a Terra data table.

        :param entity_type: the kind of entity (e.g. "sample")
        :param pandera_schema: an optional Pandera schema for the output data frame
        :return: a data frame of entities
        """

        logging.info(f"Getting {entity_type} entities from {self.workspace_name}")
        j = call_firecloud_api(
            firecloud_api.get_entities,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            etype=entity_type,
        )

        records = [{f"{entity_type}_id": x["name"], **x["attributes"]} for x in j]
        df = pd.DataFrame(records)

        if pandera_schema is None:
            return df

        return type_data_frame(df, pandera_schema)

    def upload_entities(
        self, df: pd.DataFrame, model: str = "flexible", delete_empty: bool = True
    ) -> None:
        """
        Upload a data frame of entities to a Terra data table.

        :param df: a data frame of entities
        :param model: the entity upload model to use ("flexible" is recommended over
        "firecloud")
        :param delete_empty: whether an empty cell in the data frame should clear the
        corresponding value in the Terra data table
        """

        logging.info(f"{len(df)} entities to upload to {self.workspace_name}")
        buffer = StringIO()  # store batches of TSV rows in this buffer

        for batch in batch_evenly(df, max_batch_size=500):
            logging.info(f"Upserting {len(batch)} entities to {self.workspace_name}")

            # write the latest batch of TSV rows to the emptied buffer
            buffer.seek(0)
            buffer.truncate(0)
            batch.to_csv(buffer, sep="\t", index=False)  # pyright: ignore

            call_firecloud_api(
                firecloud_api.upload_entities,
                namespace=self.workspace_namespace,
                workspace=self.workspace_name,
                entity_data=buffer.getvalue(),
                model=model,
                delete_empty=delete_empty,
            )

    def delete_entities(self, entity_type: str, entity_ids: set[str]) -> None:
        """
        Delete entities from a Terra workspace, including their related entities.

        :param entity_type: the type of entity to delete
        :param entity_ids: a set of entity IDs to delete
        """

        if len(entity_ids) == 0:
            logging.info(f"No entities in {self.workspace_name} to delete")
            return

        # get all entities
        all_entities = call_firecloud_api(
            firecloud_api.get_entities_with_type,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
        )

        for x in all_entities:
            if x["entityType"] == entity_type:
                # we're looking for entities that might *reference* the entity type
                # we're deleting, not this entity type itself, which we'll delete
                # directly later
                continue

            if "attributes" in x:
                x2 = x.copy()
                x_updated = False

                for k, v in x["attributes"].items():
                    if not (
                        isinstance(v, dict)
                        and "itemsType" in v
                        and v["itemsType"] == "EntityReference"
                    ):
                        # this is some other irrelevant entity type
                        continue

                    items = v["items"]

                    if len(items) == 0:
                        continue

                    # remove entity IDs from the join table
                    updated_items = [
                        y
                        for y in items
                        if y["entityType"] == entity_type
                        and y["entityName"] not in entity_ids
                    ]

                    if len(items) == len(updated_items):
                        continue

                    # we changed the list of referenced items
                    x2["attributes"][k]["items"] = updated_items
                    x_updated = True

                if x_updated:
                    logging.info(
                        f"Removing entities from {x2['name']} in {self.workspace_name}"
                    )

                    attribute_name = list(x2["attributes"].keys())[0]

                    # update the entity for the join table
                    _ = call_firecloud_api(
                        firecloud_api.update_entity,
                        namespace=self.workspace_namespace,
                        workspace=self.workspace_name,
                        etype=x2["entityType"],
                        ename=x2["name"],
                        updates=[
                            {
                                "op": "AddUpdateAttribute",
                                "attributeName": attribute_name,
                                "addUpdateAttribute": x2["attributes"][attribute_name],
                            }
                        ],
                    )

        # now that we've deleted related entities, delete the entities themselves
        call_firecloud_api(
            firecloud_api.delete_entities,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            json_body=[
                {"entityType": entity_type, "entityName": x} for x in entity_ids
            ],
        )

    def create_entity_set(
        self,
        entity_type: str,
        entity_ids: Iterable[str],
        entity_set_id: str | None = None,
        suffix: str | None = None,
    ) -> str:
        """
        Create a new entity set for a list of entity IDs and upload it to Terra.

        :param entity_type: the kind of entity (e.g. "sample")
        :param entity_ids: a list of entity IDs
        :param entity_set_id: an optional ID for the new entity set (if `suffix` isn't
        provided)
        :param suffix: a suffix to add to an auto-generated, timestamped, entity set ID
        (if `entity_set_id` isn't provided)
        :return: the ID of the new entity set
        """

        if entity_set_id is None:
            assert suffix is not None, (
                "suffix is required if you don't specify a entity set ID"
            )

            # generate an ID for the entity set of new entities
            entity_set_id = "_".join(
                [
                    entity_type,
                    datetime.datetime.now(datetime.UTC)
                    .isoformat(timespec="seconds")
                    .rstrip("+00:00")
                    .replace(":", "-"),
                    suffix,
                ]
            )

        elif suffix is not None:
            logging.warning(
                "An entity_set_id was provided, so the suffix option is ignored"
            )

        # construct a data frame of entity IDs for this entity set
        entity_set = pd.DataFrame(
            {f"entity:{entity_ids}_id": entity_ids}, dtype="string"
        )
        entity_set[f"entity:{entity_type}_set_id"] = entity_set_id

        logging.info(f"Creating new {entity_type} set in {self.workspace_name}")
        self.upload_entities(
            entity_set.loc[:, [f"entity:{entity_type}_set_id"]].drop_duplicates()
        )

        # construct the join/membership table between the entity set and its entities
        entity_set = entity_set.rename(
            columns={
                f"entity:{entity_type}_set_id": f"membership:{entity_type}_set_id",
                f"entity:{entity_ids}_id": entity_type,
            }
        )

        entity_set = entity_set.loc[
            :, [f"membership:{entity_type}_set_id", entity_type]
        ]

        logging.info(
            f"Adding {len(entity_set)} {entity_type} entities "
            f"to {entity_type} set {entity_set_id} in {self.workspace_name}"
        )
        self.upload_entities(entity_set)

        return entity_set_id

    def create_method_config(self, config_body: dict) -> None:
        """
        Create a Terra method config.

        :param config_body: a dictionary containing the method config
        """

        logging.info("Creating method config")
        call_firecloud_api(
            firecloud_api.create_workspace_config,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            body=config_body,
        )

        logging.info("Setting workspace config ACL")
        call_firecloud_api(
            firecloud_api.update_workspace_acl,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            acl_updates=[{"email": x, "accessLevel": "OWNER"} for x in self.owners],
        )

    def update_method_config(
        self, terra_workflow: TerraWorkflow, config_body: dict
    ) -> None:
        """
        Update a Terra method config.

        :param terra_workflow: a `TerraWorkflow` instance
        :param config_body: a dictionary containing the method config
        """

        logging.info("Update method config")
        call_firecloud_api(
            firecloud_api.update_workspace_config,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            cnamespace=terra_workflow.method_config_namespace,
            configname=terra_workflow.method_config_name,
            body=config_body,
        )

        logging.info("Setting workspace config ACL")
        call_firecloud_api(
            firecloud_api.update_workspace_acl,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            acl_updates=[{"email": x, "accessLevel": "OWNER"} for x in self.owners],
        )

    def update_workflow(
        self, terra_workflow: TerraWorkflow, n_snapshots_to_keep: int = 20
    ) -> None:
        """
        Update the Terra workflow (method and method config).

        :param terra_workflow: a `TerraWorkflow` instance
        :param n_snapshots_to_keep: the number of method snapshots to keep
        """

        # update or create the method
        snapshot = terra_workflow.update_method(self.owners)

        # assocate the method config with the latest method version
        terra_workflow.method_config["methodRepoMethod"]["methodVersion"] = snapshot[
            "snapshotId"
        ]

        logging.info("Checking for existing method config")
        # all of the `firecloud_api.*_workspace_config` methods are really operations on
        # a method config, just inside a workspace
        res = firecloud_api.get_workspace_config(
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            cnamespace=terra_workflow.method_config_namespace,
            config=terra_workflow.method_config_name,
        )

        # update or create the method config
        if res.status_code == 404:
            self.create_method_config(terra_workflow.method_config)
        else:
            self.update_method_config(terra_workflow, terra_workflow.method_config)

        # don't let old method versions accumulate
        terra_workflow.delete_old_method_snapshots(
            n_snapshots_to_keep=n_snapshots_to_keep
        )

    def check_submittable_entities(
        self,
        entity_type: str,
        entity_ids: Iterable[str],
        terra_workflow: TerraWorkflow,
        since: datetime.datetime | None = None,
        resubmit_n_times: int = 0,
        force_retry: bool = False,
    ) -> SubmittableEntities:
        """
        Check a list of entity IDs that might have been submitted to a particular
        workflow and determine whether they eligible to be (re)submitted.

        :param entity_type: the kind of entity (e.g. "sample")
        :param entity_ids: a list of entity IDs
        :param terra_workflow: a workflow to potentially create a job for
        :param since: don't collect job submissions before this `datetime`
        :param resubmit_n_times: the number of times to resubmit a failed entity to a
        job
        :param force_retry: allow resubmission even if an entity has failed more than
        `resubmit_n_times` for that workflow
        :return: a dictionary of statuses and entity IDs belong to them
        """

        # get all submissions in the workspace
        submissions = type_data_frame(
            pd.DataFrame(
                call_firecloud_api(
                    firecloud_api.list_submissions,
                    namespace=self.workspace_namespace,
                    workspace=self.workspace_name,
                )
            ),
            Submissions,
        )

        # filter submissions to those for the TerraWorkflow of interest
        subs_for_workflow = submissions.loc[
            submissions["methodConfigurationNamespace"].eq(
                terra_workflow.method_config_namespace
            )
            & submissions["methodConfigurationName"].str.startswith(
                terra_workflow.method_config_name
            ),  # Terra appends a random string to the method config used in a job
            ["submissionDate", "submissionId"],
        ]

        subs_for_workflow["submissionDate"] = pd.to_datetime(
            subs_for_workflow["submissionDate"]
        )

        if since is not None:
            # the list of submissions can grow very long and the endpoint doesn't
            # support filtering
            subs_for_workflow = subs_for_workflow.loc[
                subs_for_workflow["submissionDate"].ge(pd.Timestamp(since, tz="UTC"))
            ]

        # collect the entities submitted as part of this job and their status
        submitted_entities = []

        for sid in subs_for_workflow["submissionId"]:
            submission = call_firecloud_api(
                firecloud_api.get_submission,
                namespace=self.workspace_namespace,
                workspace=self.workspace_name,
                submission_id=sid,
            )

            for w in submission["workflows"]:
                if "workflowEntity" not in w:
                    # this workflow didn't manage to start
                    continue

                submitted_entities.append(
                    {
                        "entity_type": w["workflowEntity"]["entityType"],
                        "entity_id": w["workflowEntity"]["entityName"],
                        "status": w["status"],
                    }
                )

        # make data frame of previously submitted entities
        sub_ent_df = type_data_frame(
            pd.DataFrame(submitted_entities), SubmittedEntities
        )

        sub_ent_df = sub_ent_df.loc[
            sub_ent_df["entity_type"].eq(entity_type)
            & sub_ent_df["entity_id"].isin(list(entity_ids))
        ]

        # count number of failures per entity
        failure_counts = (
            sub_ent_df.loc[sub_ent_df["status"].eq("Failed"), "entity_id"]
            .value_counts()
            .reset_index()
        )

        if force_retry:
            # we aren't returning counts, so just zero them out to simplify logic
            failure_counts["count"] = 0

        return {
            # if an input entity ID isn't in the history, it hasn't been submitted yet
            "unsubmitted": set(
                [x for x in entity_ids if x not in sub_ent_df["entity_id"]]
            ),
            # a running entity might have state 'Queued', 'Submitted', 'Launching',
            # 'Running', or anything that's not one of the below
            "running": set(
                sub_ent_df.loc[
                    ~sub_ent_df["status"].isin(
                        ["Aborted", "Aborting", "Succeeded", "Failed"]
                    ),
                    "entity_id",
                ]
            ),
            # failed tasks are potentially retryable based on above logic
            "retryable": set(
                failure_counts.loc[
                    failure_counts["count"].le(resubmit_n_times), "entity_id"
                ]
            ),
            "failed": set(
                failure_counts.loc[
                    failure_counts["count"].gt(resubmit_n_times), "entity_id"
                ]
            ),
        }

    def get_workflow_config(self, terra_workflow: TerraWorkflow) -> dict[str, Any]:
        """
        Get the method configuration for a workflow in this workspace.

        :param terra_workflow: a `TerraWorkflow` instance
        """

        return call_firecloud_api(
            firecloud_api.get_workspace_config,
            namespace=self.workspace_namespace,
            workspace=self.workspace_name,
            cnamespace=terra_workflow.method_config_namespace,
            config=terra_workflow.method_config_name,
        )

    def submit_workflow_run(
        self, terra_workflow: TerraWorkflow, **kwargs: Unpack[TerraJobSubmissionKwargs]
    ) -> None:
        """
        Submit a run of a workflow.

        :param terra_workflow: a `TerraWorkflow` instance
        """

        logging.info(
            f"Submitting {terra_workflow.method_name} job in {self.workspace_name}"
        )
        call_firecloud_api(
            firecloud_api.create_submission,
            wnamespace=self.workspace_namespace,
            workspace=self.workspace_name,
            cnamespace=terra_workflow.method_config_namespace,
            config=terra_workflow.method_config_name,
            **kwargs,
        )

    def submit_delta_job(
        self,
        terra_workflow: TerraWorkflow,
        entity_type: str,
        entity_set_type: str,
        entity_id_col: str,
        expression: str,
        input_cols: set[str] | None = None,
        output_cols: set[str] | None = None,
        resubmit_n_times: int = 0,
        force_retry: bool = False,
        use_callcache: bool = True,
        use_reference_disks: bool = False,
        memory_retry_multiplier: float = 1.5,
        max_n_entities: int | None = None,
        dry_run: bool = False,
    ):
        """
        Identify entities in a Terra data table that need to have a workflow run on them
        by:

            1. checking for the presence of a workflow output in a data table column
            2. confirming the entity is eligible to be submitted in a job by checking
               for previous submissions of that same entity to the workflow

        :param terra_workflow: a TerraWorkflow instance for the method
        :param entity_type: the name of the Terra entity type
        :param entity_set_type: the name of the Terra entity set type for `entity_type`
        :param entity_id_col: the name of the ID column for the entity type
        :param expression: the entity type expression (e.g. "this.samples")
        :param input_cols: the set of column names that must all be present in the
        entity type in order for an entity to be submittable
        :param output_cols: the set of column names that must all be missing in the
        entity type in order for an entity to be submittable
        :param resubmit_n_times: the number of times to resubmit an entity in the event
        it has failed in the past
        :param force_retry: whether to retry even if `resubmit_n_times` has been reached
        :param use_callcache: whether to use call caching
        :param use_reference_disks: whether to use reference disks
        :param memory_retry_multiplier: a multiplier for retrying with more memory
        :param max_n_entities: submit at most this many entities (random sample)
        :param dry_run: whether to skip updates to external data stores
        """

        # get the method config for this workflow in this workspace
        workflow_config = self.get_workflow_config(terra_workflow)

        assert not workflow_config["deleted"]
        assert workflow_config["rootEntityType"] == entity_type

        # identify columns in data table used for input/output if not provided
        if input_cols is None:
            input_cols = {
                v[5:]
                for k, v in workflow_config["inputs"].items()
                if v.startswith("this.")
            }

        if output_cols is None:
            output_cols = {
                v[5:]
                for k, v in workflow_config["outputs"].items()
                if v.startswith("this.")
            }

        # get the entities for this workflow entity type
        entities = self.get_entities(entity_type)

        # ensure columns exist to check for populated values
        for c in input_cols.union(output_cols):
            if c not in entities.columns:
                entities[c] = pd.NA
            else:
                entities[c] = entities[c].replace({"": pd.NA})

        # identify entities that have all required inputs but no outputs
        entities_todo = entities.loc[
            entities[list(input_cols)].notna().all(axis=1)
            & entities[list(output_cols)].isna().all(axis=1)
        ]

        if len(entities_todo) == 0:
            logging.info(f"No {entity_type}s to run {terra_workflow.method_name} for")
            return

        # get statuses of submitted workflows for the entitites
        submittable_entities = self.check_submittable_entities(
            entity_type,
            entity_ids=entities_todo[entity_id_col],
            terra_workflow=terra_workflow,
            resubmit_n_times=resubmit_n_times,
            force_retry=force_retry,
        )

        logging.info(f"Submittable entities: {submittable_entities}")

        if len(submittable_entities["failed"]) > 0:
            raise RuntimeError("Some entities have failed too many times")

        # don't submit jobs for entities that are currently running, completed, or
        # failed too many times
        entities_todo = entities_todo.loc[
            entities_todo[entity_id_col].isin(
                list(
                    submittable_entities["unsubmitted"]
                    .union(submittable_entities["retryable"])
                    .difference(submittable_entities["running"])
                )
            )
        ]

        if len(entities_todo) == 0:
            logging.info("No entities to submit")
            return
        elif max_n_entities is not None and len(entities_todo) > max_n_entities:
            logging.info(f"Sampling {max_n_entities} of {len(entities_todo)} entities")
            entities_todo = entities_todo.sample(n=max_n_entities)

        if dry_run:
            logging.info(f"(skipping) Submitting {terra_workflow.method_name} job")
            return

        entity_set_id = self.create_entity_set(
            entity_type,
            entity_ids=entities_todo[entity_id_col],
            suffix=terra_workflow.method_name,
        )

        self.submit_workflow_run(
            terra_workflow=terra_workflow,
            entity=entity_set_id,
            etype=entity_set_type,
            expression=expression,
            use_callcache=use_callcache,
            use_reference_disks=use_reference_disks,
            memory_retry_multiplier=memory_retry_multiplier,
        )

    def collect_workflow_outputs(
        self, since: datetime.datetime | None = None
    ) -> list[TaskResult]:
        """
        Collect workflow outputs and metadata about the jobs that ran them.

        :param since: don't collect outputs for job submissions before this `datetime`
        :return: a list of Gumbo `task_result` objects to insert
        """

        logging.info(f"Getting previous job submissions in {self.workspace_name}")
        submissions = type_data_frame(
            pd.DataFrame(
                call_firecloud_api(
                    firecloud_api.list_submissions,
                    namespace=self.workspace_namespace,
                    workspace=self.workspace_name,
                )
            ),
            Submissions,
        )

        submissions = pd_flatten(submissions).convert_dtypes()

        submissions["submissionDate"] = pd.to_datetime(submissions["submissionDate"])

        if since is not None:
            # the list of submissions can grow very long and the endpoint doesn't
            # support filtering
            submissions = submissions.loc[
                submissions["submissionDate"].ge(pd.Timestamp(since, tz="UTC"))
            ]

        # only collect outputs from submissions with successful workflow runs
        submissions = submissions.loc[submissions["workflowStatuses__Succeeded"].gt(0)]

        outputs = []

        for _, s in submissions.iterrows():
            sid = s["submissionId"]

            # start constructing a common output object to be used for `task_result`
            # inserts
            base_o = TaskResult(
                terra_method_config_name=str(s["methodConfigurationName"]),
                terra_method_config_namespace=str(s["methodConfigurationNamespace"]),
                terra_submission_id=str(sid),
                terra_workspace_name=self.workspace_name,
                terra_workspace_namespace=self.workspace_namespace,
            )

            # get the workflows for this job submission (often there is only 1)
            logging.info(f"Getting workflows for job submission {sid}")
            submission = call_firecloud_api(
                firecloud_api.get_submission,
                namespace=self.workspace_namespace,
                workspace=self.workspace_name,
                submission_id=sid,
            )

            for w in submission["workflows"]:
                if "workflowId" not in w:
                    # this workflow didn't manage to start
                    continue

                if "workflowEntity" not in w:
                    # this workflow didn't manage to start
                    continue

                wid = w["workflowId"]
                base_o.terra_workflow_id = wid
                base_o.terra_entity_name = w["workflowEntity"]["entityName"]
                base_o.terra_entity_type = w["workflowEntity"]["entityType"]

                base_o.completed_at = pd.Timestamp(
                    ts_input=w["statusLastChangedDate"], tz="UTC"
                ).isoformat()  # pyright: ignore

                logging.info(f"Getting workflow {wid} metadata")
                wmd = call_firecloud_api(
                    firecloud_api.get_workflow_metadata,
                    namespace=self.workspace_namespace,
                    workspace=self.workspace_name,
                    submission_id=sid,
                    workflow_id=wid,
                    include_key=[
                        "inputs",
                        "labels",
                        "outputs",
                        "status",
                        "workflowName",
                        "workflowRoot",
                    ],
                )

                if wmd["status"] != "Succeeded":
                    continue

                base_o.workflow_name = wmd["workflowName"]
                base_o.terra_workflow_inputs = wmd["inputs"]
                base_o.terra_workflow_root_dir = wmd["workflowRoot"]
                base_o.terra_workspace_id = wmd["labels"]["workspace-id"]

                # iterate through the outputs and make a `task_result` object for each
                for label, output in wmd["outputs"].items():
                    # all attributes up to now have been invariant between outputs
                    o = base_o.model_copy()

                    # the workflow outputs are named like `workflow_name.output_name`
                    o.label = label.rsplit(".", maxsplit=1)[-1]

                    if isinstance(output, list):
                        # we shouldn't be writing any workflows that output lists, so
                        # these would just be intermediate or legacy outputs
                        continue
                    elif isinstance(output, str):
                        if output.startswith("gs://"):
                            o.url = output
                            o.format = pathlib.Path(o.url).suffix[1:].upper()
                        else:
                            # it's a plain string
                            o.value = {"value": output}
                    else:
                        # it's a number, bool, or something else
                        o.value = {"value": output}

                    outputs.append(o)

        return outputs
